@@ -195,7 +195,7 @@ enum ce { CE_DEBUG, CE_CONT, CE_NOTE, CE_WARN, CE_ALERT, CE_PANIC };
  */
 #define prandom_u32()		0
 
-#define PAGE_CACHE_SIZE		getpagesize()
+#define PAGE_SIZE		getpagesize()
 
 static inline int __do_div(unsigned long long *n, unsigned base)
 {
@@ -361,7 +361,6 @@ roundup_64(__uint64_t x, __uint32_t y)
 #define XFS_MOUNT_RDONLY		0	/* ignored in userspace */
 
 
-#define _xfs_trans_alloc(mp, type, f)	libxfs_trans_alloc(mp, type)
 #define xfs_trans_get_block_res(tp)	1
 #define xfs_trans_set_sync(tp)		((void) 0)
 #define xfs_trans_ordered_buf(tp, bp)	((void) 0)
@@ -371,6 +370,7 @@ roundup_64(__uint64_t x, __uint32_t y)
 #define xfs_trans_buf_set_type(tp, bp, t)	({	\
 	int __t = (t);					\
 	__t = __t; /* no set-but-unused warning */	\
+	tp = tp;  /* no set-but-unused warning */	\
 })
 
 #define xfs_trans_buf_copy_type(dbp, sbp)
@@ -474,8 +474,7 @@ struct xfs_buftarg;
 int xfs_attr_rmtval_get(struct xfs_da_args *);
 
 /* xfs_bmap.c */
-void xfs_bmap_del_free(struct xfs_bmap_free *, struct xfs_bmap_free_item *,
-			struct xfs_bmap_free_item *);
+void xfs_bmap_del_free(struct xfs_bmap_free *, struct xfs_bmap_free_item *);
 
 /* xfs_mount.c */
 int xfs_initialize_perag_data(struct xfs_mount *, xfs_agnumber_t);
